from lxml import etree
import os
from glob import glob
from copy import deepcopy
from .string import remove_whitespace, prettify
from IPython.display import display, Markdown
from pathlib import Path
from lxml import etree

# TEI namespace
tei_namespace = "http://www.tei-c.org/ns/1.0"
ns = {"tei": tei_namespace }

def add_number_to_bibl(dir_path):
    """written by GPT-4"""
    for file_path in glob(f'{dir_path}/*.xml'):
        print(f' - Processing {file_path}')
        n = 1
        tree = etree.parse(file_path)
        for element in tree.xpath('*//tei:bibl', namespaces=ns):
            element.attrib['n'] = str(n)
            n += 1
        pretty_tree = etree.tostring(tree, pretty_print=True, encoding='utf-8')
        with open(file_path, 'wb') as file:
            file.write(pretty_tree)

def remove_encoding_declaration(xml_string):
    return xml_string.replace('<?xml version="1.0" encoding="UTF-8"?>', '')

def create_gold_standard(bibl_content, biblstruct_content, verbose=False):
    # TEI namespace
    tei_namespace = "http://www.tei-c.org/ns/1.0"
    ns = {"tei": tei_namespace }

    # get source trees and elements
    bibl_tree = etree.fromstring(remove_encoding_declaration(bibl_content))
    bibl_struct_tree = etree.fromstring(remove_encoding_declaration(biblstruct_content))

    # create output document
    output_tree = etree.Element("TEI", {'xmlns':tei_namespace})
    tei_header = bibl_struct_tree.xpath('./tei:teiHeader', namespaces=ns)[0]
    output_tree.append(deepcopy(tei_header))
    standoff = etree.SubElement(output_tree, "standOff")

    # keep track of parent, which can be "note" or "listBibl"
    current_source_parent = None
    # keep track of target listBibl node
    current_target_parent = None

    process_log = []

    # iterate over bibl elements
    for bibl_element in bibl_tree.xpath('//tei:bibl', namespaces=ns):

        # the number of the bibl element in the document is required
        if 'n' not in bibl_element.attrib:
            raise RuntimeError(f"Missing 'n' attribute for bibl element: {bibl_element}")
        bibl_n = bibl_element.attrib['n']

        # node parent
        source_parent = bibl_element.getparent()
        # if the parent changes, create a new target parent
        if source_parent != current_source_parent:
            current_source_parent = source_parent
            parent_tag = etree.QName(current_source_parent).localname
            parent_type = current_source_parent.attrib['type']

            if parent_type == "footnote" or parent_type == "inText":
                source_node = source_parent
            else:
                source_node = bibl_element
                # invalidate the pointer to the current source parent to create a new listBibl the next time
                current_source_parent = None

            attrs = {
                "type": parent_type,
            }

            footnote_number = None
            if parent_type == "footnote":
                footnote_number = attrs['n'] = source_node.attrib['n']

                # create mew listBibl element for output element
            current_target_parent = etree.SubElement(standoff, "listBibl", attrs)

            # create <desc> child to store the full input string
            input_full = etree.SubElement(current_target_parent, "desc", {'type': 'input-full'})

            # get content for it
            child_xml = etree.tostring(source_node, method="text", encoding='utf-8').decode()
            xml_content_text = remove_whitespace(child_xml)
            if footnote_number:
                xml_content_text = f'{footnote_number} {xml_content_text}'
            input_full.text = xml_content_text

            # create <desc> child to store the segmented input string
            input_segmented = etree.SubElement(current_target_parent, "desc", {'type': 'input-segmented'})

            # if bibliography, the segmented input string is the same as the raw
            if parent_type == "bibliography":
                etree.SubElement(input_segmented, "bibl").text = xml_content_text

            else:
                # Generate the high-level segments of the input string
                # For footnotes and in-text references, these are usually one or more <bibl> elements;
                # for bibliographies, there is one <bibl> element per <listBibl> element.
                # As there aren't any suitable other legal children elements of <desc>, we use <desc> and <bibl> to
                # segment the raw reference strings:
                # - we create a child <desc> element as direct child of the target <desc> element for each
                #   non-bibliographic element within the source <bibl> so that we don't have a nested structure
                # - we create a <bibl> element with the raw string content of the source <bibl> without the <seg type="comment"> parts
                for elem in source_node:

                    last_element = None
                    last_element_parts = None

                    for segment in elem:
                        segment_text_content = etree.tostring(segment, method="text", encoding='utf-8').decode().strip()

                        if 'type' in segment.attrib:
                            if segment.attrib['type'] == 'comment' or segment.attrib['type'] == 'signal':
                                comment_child = etree.SubElement(input_segmented, "desc")
                                comment_child.text = segment_text_content
                                last_element = None
                        else:
                            if last_element is None:
                                last_element = etree.SubElement(input_segmented, "bibl")
                                last_element_parts = [segment_text_content]
                            else:
                                last_element_parts.append(segment_text_content)
                            last_element.text = remove_whitespace(" ".join(last_element_parts))

                if footnote_number:
                    input_segmented[0].text = f'{footnote_number} {input_segmented[0].text}'

            if verbose:
                process_log.append(f' - Analyzing "{xml_content_text[:200]} [...]"')

        # find correspondent biblStruct data
        bibl_struct_matches = bibl_struct_tree.xpath(f"//tei:biblStruct[@n='{bibl_n}']", namespaces=ns)
        if len(bibl_struct_matches) > 0:
            # found it - make a copy and remove unneeded attributes
            bibl_struct_copy = deepcopy(bibl_struct_matches[0])
            if bibl_struct_copy.attrib['source']:
                del bibl_struct_copy.attrib['source']
            # add it to target listBibl
            current_target_parent.append(bibl_struct_copy)
        else:
            display(Markdown("\n".join(process_log)))
            raise RuntimeError(f"Could not find matching biblStruct element with id '{bibl_n}'.")

    display(Markdown("\n".join(process_log)))
    return (etree.tostring(output_tree)).decode()


def create_all_gold_standards(bibl_dir, biblstruct_dir, biblstruct_gold_dir, verbose=False):
    for file_path in glob(f'{bibl_dir}/*.xml'):
        file_id = os.path.basename(file_path).replace(".xml", "")

        bibl_path = file_path
        biblstruct_path = f'{biblstruct_dir}/{file_id}.biblstruct.xml'
        biblstruct_gs_path = f'{biblstruct_gold_dir}/{file_id}.xml'

        # log
        md_lines = [f'### Processing {file_id}']
        md_lines.append(f'Files: [TEI/bibl]({Path(bibl_path).as_posix()}) | [TEI/biblStruct]({Path(biblstruct_path).as_posix()}) | [TEI/biblStruct Gold Standard]({Path(biblstruct_gs_path).as_posix()})')
        display(Markdown("\n".join(md_lines)))

        with (open(bibl_path, 'r', encoding='utf-8') as bibl_file,
              open(biblstruct_path, 'r', encoding='utf-8') as biblStruct_file):
            bibl_content = bibl_file.read()
            biblStruct_content = biblStruct_file.read()

        output_data = create_gold_standard(bibl_content, biblStruct_content, verbose=verbose)

        with open(biblstruct_gs_path, 'w', encoding='utf-8') as output_file:
            output_file.write(output_data)